///////////////////////////////////////////////////////////////////////////////////////////////////////////
// PaletteRotationDemo -	Written by Robert Mooney (rmooney@iss.net)
//							Image by Rafal Niedzialkowski (rniedzialkowski@iss.net), used with permission.
//
// This code may be redistributed and reused freely, assuming the author's name remains intact,
// and credit is given where appropriate.  Please contact Rafal Niedzialkowski for information
// regarding redistribution of his image.
//
// (c) Copyright 2000 Against Our Will, Inc., All rights reserved.
///////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <windows.h>
#include <windowsx.h>
#include <iostream.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ddraw.h>


////////////////////
// Misc definitions
////////////////////

#define DEFAULT_SLEEP_RATE	10				// Default 10 msec between palette swaps
#define SCREEN_WIDTH		1024			// Screen width
#define SCREEN_HEIGHT		768				// Screen height
#define SCREEN_BPP			8				// Screen Bits Per Pixel

#define BITMAP_MAGIC		0x4d42			// Bitmap magic number


//////////////////////////
// Local bitmap structure
//////////////////////////

typedef struct
{
	BITMAPFILEHEADER	bmHeader;		// The header
	BITMAPINFOHEADER	bmInfoHeader;	// The info header
	PALETTEENTRY		bmPalette[256];	// The 256 color palette
	UCHAR				*ucImage;		// The image data

} BITMAPSTRUCT;


/////////////////////////////////////////////////////////
// LoadBitmap - Load bitmap data from the file specified
/////////////////////////////////////////////////////////
HRESULT
LoadBitmap(void *pvFilename, void *pvBitmapStruct)
{
	INT iFilehandle;												// Bitmap filehandle
	BITMAPSTRUCT *bsBitmap	= (BITMAPSTRUCT *) pvBitmapStruct;		// Bitmap structure
	CHAR *szFilename		= (char *) pvFilename;					// Bitmap filename


	///////////////////////////
	// Open the specified file
	///////////////////////////
	if (!(iFilehandle = _open(szFilename, _O_RDONLY | _O_BINARY)))
	{
		return E_INVALIDARG;
	}


	//////////////////////////
	// Read the bitmap header
	//////////////////////////
	if (_read(iFilehandle, (void *) &bsBitmap->bmHeader, sizeof(bsBitmap->bmHeader)) < 0)
	{
		return E_FAIL;
	}

	// Make sure we're dealing with a bitmap
	if (bsBitmap->bmHeader.bfType != BITMAP_MAGIC)
	{
		return E_NOTIMPL;
	}

	
	///////////////////////////////
	// Read the bitmap info header
	///////////////////////////////
	if (_read(iFilehandle, (void *) &bsBitmap->bmInfoHeader, sizeof(bsBitmap->bmInfoHeader)) < 0)
	{
		return E_FAIL;
	}

	// Make sure it's an 8-bit bitmap.  Just for now.
	if (bsBitmap->bmInfoHeader.biBitCount != 8)
	{
		return E_NOTIMPL;
	}


	////////////////////////////////
	// Read and convert the palette
	////////////////////////////////

	// Read the palette
	if (_read(iFilehandle, (void *) &bsBitmap->bmPalette, sizeof(bsBitmap->bmPalette)) < 0)
	{
		return E_FAIL;
	}

	LONG lCount;			// Variables used below: counter and temp color
	BYTE bTempColor;

	// Convert the RGBQUADs to PALETTEENTRYs
	for (lCount = 1; lCount < 256; ++lCount)
	{
		// Swap the blue and red values (BGR -> RGB)
		bTempColor = bsBitmap->bmPalette[lCount].peBlue;
		bsBitmap->bmPalette[lCount].peBlue = bsBitmap->bmPalette[lCount].peRed;
		bsBitmap->bmPalette[lCount].peRed  = bTempColor;

		// Don't match colors in this palette to Windows system colors
		bsBitmap->bmPalette->peFlags = PC_NOCOLLAPSE;
	}


	///////////////////////
	// Read the image data
	///////////////////////

	// Allocate memory for the image data
	if (!(bsBitmap->ucImage = (UCHAR *) calloc(1, bsBitmap->bmInfoHeader.biSizeImage)))
	{
		return E_OUTOFMEMORY;
	}

	// Seek to the start of the image data
	if (_lseek(iFilehandle, -(long)bsBitmap->bmInfoHeader.biSizeImage, SEEK_END) < 0)
	{
		return E_FAIL;
	}

	// Read the image data
	if (_read(iFilehandle, (void *) bsBitmap->ucImage, 
		bsBitmap->bmInfoHeader.biSizeImage) != (INT) bsBitmap->bmInfoHeader.biSizeImage)
	{
		return E_FAIL;
	}

	// Close the bitmap file
	_close(iFilehandle);


	//////////////////
	// Flip the image
	//////////////////

	UCHAR	*ucNewImage;		// Fixed copy of the image data
	ULONG	ulBytesPerRow;		// Bytes per row of image data

	//  # bytes == width * bytes/pixel * height
	ulBytesPerRow = bsBitmap->bmInfoHeader.biWidth * (bsBitmap->bmInfoHeader.biBitCount / 8);

	// Allocated memory for the "fixed" image -- width x height
	if (!(ucNewImage = (UCHAR *) calloc(1, ulBytesPerRow * bsBitmap->bmInfoHeader.biHeight)))
	{
		return E_OUTOFMEMORY;
	}

	// Flip the image since it's stored upside down
	for (lCount = 0; lCount < bsBitmap->bmInfoHeader.biHeight; ++lCount)
	{
		memcpy(
			// Copy to the last row backward
			&ucNewImage[(bsBitmap->bmInfoHeader.biHeight - lCount - 1) * ulBytesPerRow],
			// Copy from the first row forward
			&bsBitmap->ucImage[lCount * ulBytesPerRow],
			// Copy the number of bytes in a row of data
			ulBytesPerRow
			);
	}

	// Free up the memory previously allocated
	free(bsBitmap->ucImage);


	/////////////////////////////////////////////////////
	// Point the image buffer to the new "correct" image
	/////////////////////////////////////////////////////
	bsBitmap->ucImage = ucNewImage;

	return S_OK;
}


///////////////////////////////////////////////////////////////////
// UnloadBitmap - Free memory previously allocated for bitmap data 
///////////////////////////////////////////////////////////////////
HRESULT
UnloadBitmap(void *pvBitmapStruct)
{
	BITMAPSTRUCT *bsBitmap = (BITMAPSTRUCT *) pvBitmapStruct;

	// If the memory hasn't been free()'d yet, free it.
	if (bsBitmap->ucImage)
	{
		free(bsBitmap->ucImage);
		bsBitmap->ucImage = NULL;
	}

	return S_OK;
}

////////////////////////////////////////////////////////////////////////////////////////
// NraiseN - Simple function to perform x ^ y (x to the power of y), as on a calculator
////////////////////////////////////////////////////////////////////////////////////////
ULONG
NraiseN(LONG ln0, LONG ln1)
{
	ULONG ulAnswer = ln0;

	// Sorry, no negative numbers/floats.
	if ((ln0 < 0) || (ln1 < 0))  return 0;

	while (--ln1 > 0)  ulAnswer *= ln0;

	return ulAnswer;
}


////////////////////////////////////////////////
// EventHandler - Windows event/message handler
////////////////////////////////////////////////
LRESULT CALLBACK
EventHandler(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{

	switch(msg)
	{

	// Window has been created
	case WM_CREATE:
		return 0;
		
	// Window needs repainting
	case WM_PAINT:

		// Validate the window
		ValidateRect(hWnd, NULL);

		return 0;

	// Window is closing
	case WM_DESTROY:

		// Post a QUIT message to the event queue
		PostQuitMessage(0);		
		return 0;

	// Have the default message handler take care of it
	default:
		break;
	}

	// Dispatch to the default message handler
	return ( DefWindowProc(hWnd, msg, wParam, lParam) );
}


//////////////////////////
// WinMain - Main program
//////////////////////////
int WINAPI
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{

	///////////////////////////////
	// Initialize the window class
	///////////////////////////////
	WNDCLASS wcWindowClass;

	memset(&wcWindowClass, 0, sizeof(wcWindowClass));

	// Set the window name
	wcWindowClass.lpszClassName = "PALETTEROTDEMO";

	// Set the application instance
	wcWindowClass.hInstance = hInstance;

	// Set the event handler
	wcWindowClass.lpfnWndProc = EventHandler;

	// Set the window style
	wcWindowClass.style =	CS_DBLCLKS	|			// Send message for double click
							CS_HREDRAW	|			// Send message when a horizontal redraw is needed
							CS_VREDRAW	|			// Send message when a vertical redraw is needed
							CS_OWNDC;				// Give us our own device context

	// Set the application icon and cursor
	wcWindowClass.hIcon		= LoadIcon   (NULL, IDI_APPLICATION);
	wcWindowClass.hCursor	= LoadCursor (NULL, IDC_ARROW);

	// Don't show the cursor
	ShowCursor(FALSE);

	// Set the background brush
	wcWindowClass.hbrBackground = (HBRUSH) GetStockObject (BLACK_BRUSH);


	/////////////////////////////
	// Register the window class
	/////////////////////////////

	if (RegisterClass(&wcWindowClass) == 0)
	{
		cout << "An error occurred while registering the window class\n";
		return 1;
	}


	/////////////////////
	// Create the window
	/////////////////////

	HWND hwMainWindow = CreateWindow(
		"PALETTEROTDEMO",				// The class name
		"Palette Rotation Demo",		// The window title
		WS_POPUP | WS_VISIBLE,			// Create a popup window with no border, and make it visible to start
		CW_USEDEFAULT, CW_USEDEFAULT,	// Default start position
		CW_USEDEFAULT, CW_USEDEFAULT,	// Default window size
		NULL,							// Handle to parent window
		NULL,							// Handle to menu
		hInstance,						// WinMain() instance
		NULL							// Additional parameters passed to the window when created
		);

	if (hwMainWindow == NULL)
	{
		cout << "An error occurred while creating the window\n";
		return 1;
	}


	/////////////////////////
	// Initialize DirectDraw
	/////////////////////////
	LPDIRECTDRAW pDirectDraw;
	HRESULT hr;

	// Create a DirectDraw object, using the active display
	hr = DirectDrawCreate(NULL, &pDirectDraw, NULL);

	if (FAILED(hr))
	{
		ShowCursor(TRUE);
		MessageBox(hwMainWindow, "An error occurred while creating the DirectDraw object.", "Error", MB_OK);
		ShowCursor(FALSE);

		return 1;
	}

	// Set the window behavior
	hr = pDirectDraw->SetCooperativeLevel(
		hwMainWindow,											// Window handle
		DDSCL_ALLOWREBOOT | DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN	// Allow reboots, set fullscreen mode
		);

	if (FAILED(hr))
	{
		ShowCursor(TRUE);
		MessageBox(hwMainWindow, "An error occurred while setting the cooperative level.", "Error", MB_OK);
		ShowCursor(FALSE);

		return 1;
	}

	// Set screen dimensions
	hr = pDirectDraw->SetDisplayMode(SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_BPP);

	if (FAILED(hr))
	{
		ShowCursor(TRUE);
		MessageBox(hwMainWindow, "An error occurred while setting the display mode.", "Error", MB_OK);
		ShowCursor(FALSE);

		return 1;
	}


	/////////////////////////////////
	// Initialize DirectDraw surface
	/////////////////////////////////
	DDSURFACEDESC ddSurfaceDesc;

	// Initialize the primary surface structure
	ddSurfaceDesc.dwSize			= sizeof(ddSurfaceDesc);	// Size of the structure
	ddSurfaceDesc.dwFlags			= DDSD_CAPS;				// DDSCAPS member is valid
	ddSurfaceDesc.ddsCaps.dwCaps	= DDSCAPS_PRIMARYSURFACE;	// Specify the primary surface


	LPDIRECTDRAWSURFACE pDirectDrawSurface;

	// Create a surface for the DirectDraw object
	hr = pDirectDraw->CreateSurface(
		&ddSurfaceDesc,				// DirectDraw surface description
		&pDirectDrawSurface,		// [out] Pointer to DirectDraw surface
		NULL						// Reserved
		);

	if (FAILED(hr))
	{
		ShowCursor(TRUE);
		MessageBox(hwMainWindow, "An error occurred while creating the primary surface.", "Error", MB_OK);
		ShowCursor(FALSE);

		return 1;
	}


	/////////////////////////
	// Display the About box
	/////////////////////////

	// ** If this isn't displayed, the modifications made to the buffer
	//    below won't show up... WHY?
	// ** Message boxes cannot be displayed when the surface is locked
	ShowCursor(TRUE);
	MessageBox(hwMainWindow, "Controls:\n\n" \

	                         "Up arrow increases the palette pause time\n" \
		                     "Down arrow decreases the palette pause time\n" \
							 "Right arrow moves the palette from left to right\n" \
							 "Left arrow moves the palette from right to left\n" \
                             "ESC exits the program\n\n" \

							 "Written by Robert Mooney (rmooney@iss.net)\n" \
							 "Image by Rafal Niedzialkowski (rniedzialkowski@iss.net) <-- this guy is elite\n\n" \

							 "Code (c) Copyright 2000 Against Our Will, Inc., All rights reserved.",
							 "Control and About information...", MB_OK);
	ShowCursor(FALSE);


	///////////////////
	// Load the bitmap
	///////////////////

	BITMAPSTRUCT bsBitmap;

	hr = LoadBitmap("image.bmp", &bsBitmap);

	if (FAILED(hr))
	{
		ShowCursor(TRUE);
		MessageBox(hwMainWindow, "An error occurred while loading the background image.", "Error", MB_OK);
		ShowCursor(FALSE);

		return 1;
	}


	//////////////////////////////////////////////////////////////////////////
	// Create and set a DirectDraw palette compatible with the bitmap palette
	//////////////////////////////////////////////////////////////////////////
	LPDIRECTDRAWPALETTE pDirectDrawPalette;

	hr = pDirectDraw->CreatePalette(DDPCAPS_8BIT | DDPCAPS_INITIALIZE | DDPCAPS_ALLOW256, bsBitmap.bmPalette, &pDirectDrawPalette, NULL);

	if (FAILED(hr))
	{
		ShowCursor(TRUE);
		MessageBox(hwMainWindow, "An error occurred while creating the palette.", "Error", MB_OK);
		ShowCursor(FALSE);

		return 1;
	}

	// Set the palette for this surface
	hr = pDirectDrawSurface->SetPalette(pDirectDrawPalette);

	if (FAILED(hr))
	{
		ShowCursor(TRUE);
		MessageBox(hwMainWindow, "An error occurred while setting the palette.", "Error", MB_OK);
		ShowCursor(FALSE);

		return 1;
	}


	///////////////////////////////////////////////////////////////////
	// Reinitialize the primary surface structure and lock the surface
	///////////////////////////////////////////////////////////////////

	memset(&ddSurfaceDesc, 0, sizeof(ddSurfaceDesc));
	ddSurfaceDesc.dwSize = sizeof(ddSurfaceDesc);

	hr = pDirectDrawSurface->Lock(
		NULL,										// Lock the entire surface
		&ddSurfaceDesc,								// Surface description struct
		DDLOCK_SURFACEMEMORYPTR | DDLOCK_WAIT,		// Return a pointer to the surface, block if necessary
		NULL										// Unused
		);

	if (FAILED(hr))
	{
		ShowCursor(TRUE);
		MessageBox(hwMainWindow, "An error occurred while locking the palette.", "Error", MB_OK);
		ShowCursor(FALSE);

		return 1;
	}


	////////////////////////////////////////////////////////////////////////////////////
	// Get the current surface buffer and copy the bitmap data onto the primary surface
	////////////////////////////////////////////////////////////////////////////////////

	ULONG ulY;													// Y coordinate of the screen
	UCHAR *ucVideoBuffer = (UCHAR *) ddSurfaceDesc.lpSurface;	// Pointer to the surface buffer
	
	// Copy the image to the primary surface
	for (ulY = 0; ulY < SCREEN_WIDTH; ++ulY)
	{
		// This will only work properly if both images have the same color depth.  
		//
		// If the source image is larger than the destination surface, the image will be partially
		// displayed (with no other distortion).
		//
		memcpy(
			// Copy into each line of the primary buffer (lPitch == # bytes per horizontal line)
			&ucVideoBuffer[ddSurfaceDesc.lPitch * ulY],

			// Copy from each line of the bitmap (Current position == # bytes per horizontal line * Y coordinate)
			&bsBitmap.ucImage[bsBitmap.bmInfoHeader.biWidth * (bsBitmap.bmInfoHeader.biBitCount / 8) * ulY],

			// Copy as many bytes as will fit into a line of the primary buffer.
			ddSurfaceDesc.lPitch
			);
	}


	///////////////////////
	// Unlock the surface
	///////////////////////
	hr = pDirectDrawSurface->Unlock(ddSurfaceDesc.lpSurface);

	if (FAILED(hr))
	{
		ShowCursor(TRUE);
		MessageBox(hwMainWindow, "An error occurred while unlocking the palette.", "Error", MB_OK);
		ShowCursor(FALSE);

		return 1;
	}


	///////////////////////////////
	// GDI message processing loop
	///////////////////////////////
	MSG				msgMainWindow;							// Message from the main window
	HDC				hdc;									// Surface device context
	BOOL			bPaletteDirectionRightToLeft = TRUE;	// Direction of the palette
	CHAR			szTextBuffer[50];						// Buffer which stores text to be displayed on-screen
	USHORT			usCounter = 0,							// Loop counter (number of processing loops since last palette change)
					usSleepRate = DEFAULT_SLEEP_RATE;		// Current sleep rate (number of processing loops before palette change should take place)
	ULONG			ulPaletteColors;						// Number of colors in the screen palette
	PALETTEENTRY	peCurrentPalette[256];					// Current work palette (where we do our swaps)


	// Figure out how many colors we have in our palette (2 ^ SCREEN_BPP)
	ulPaletteColors = NraiseN(2, SCREEN_BPP);

	if (ulPaletteColors == 0)
	{
		ShowCursor(TRUE);
		MessageBox(hwMainWindow, "An error occurred while determining the palette size.", "Error", MB_OK);
		ShowCursor(FALSE);

		return 1;
	}

	// Setup stuff for the text (sleep rate) display
	pDirectDrawSurface->GetDC(&hdc);
	SetBkColor(hdc, RGB(0,0,0));
	SetTextColor(hdc, RGB(255,255,255));

	while (1)
	{
		
		/////////////////////////////
		// Check for window messages
		/////////////////////////////

		if (PeekMessage(&msgMainWindow, NULL, 0, 0, PM_REMOVE))
		{
			// Check for QUIT
			if (msgMainWindow.message == WM_QUIT)  break;

			// Translate accelerator keys
			TranslateMessage(&msgMainWindow);

			// Dispatch the message to the event handler
			DispatchMessage(&msgMainWindow);
		}


		/////////////////////
		// Process user keys
		/////////////////////
		
		// If UP arrow was pressed, reduce the pause time between palette switches
		if (GetAsyncKeyState(VK_DOWN))
		{
			// No less than 0 milliseconds
			if (usSleepRate > 0)
			{
				usSleepRate -= 1;		// Decrease pause time by 1 msec
			}
		}

		// If DOWN arrow was pressed, increase the pause time between palette switches
		if (GetAsyncKeyState(VK_UP))
		{
			// No more than 100 milliseconds
			if (usSleepRate < 100)
			{
				usSleepRate += 1;		// Increase pause time by 1 msec
			}
		}

		// If LEFT arrow was pressed, move the palette from right to left
		if (GetAsyncKeyState(VK_LEFT))
		{
			bPaletteDirectionRightToLeft = TRUE;
		}

		// If RIGHT arrow was pressed, move the palette from left to right
		if (GetAsyncKeyState(VK_RIGHT))
		{
			bPaletteDirectionRightToLeft = FALSE;
		}
		
		// If ESC was pressed, exit
		if (GetAsyncKeyState(VK_ESCAPE))
		{
			PostMessage(hwMainWindow, WM_DESTROY, NULL, NULL);
			break;
		}


		////////////////////////////////////////////////////////////
		// Rotate the palette, display the rotation rate, and sleep
		////////////////////////////////////////////////////////////

		// If the counter is greater than or equal to the sleep rate, 
		// it's time to swap the palette
		if (usCounter >= usSleepRate)
		{
			// Reset the counter
			usCounter = 0;

			// Get the current palette
			pDirectDrawPalette->GetEntries(0, 0, ulPaletteColors, peCurrentPalette);

			// Rotate the palette from right to left
			if (bPaletteDirectionRightToLeft)
			{
				pDirectDrawPalette->SetEntries(0, 0, ulPaletteColors - 1, &peCurrentPalette[1]);
				pDirectDrawPalette->SetEntries(0, 255, 1, &peCurrentPalette[0]);
			}

			// Rotate the palette from left to right
			else
			{
				pDirectDrawPalette->SetEntries(0, 1, ulPaletteColors - 1, &peCurrentPalette[0]);
				pDirectDrawPalette->SetEntries(0, 0, 1, &peCurrentPalette[255]);
			}
		}


		// Display the sleep rate
		_snprintf(szTextBuffer, sizeof(szTextBuffer), "Sn00ze rate: %-5d", usSleepRate);
		TextOut(hdc, 0, 0, szTextBuffer, strlen(szTextBuffer));

		// If the sleep rate is greater than 0, sleep
		if (usSleepRate > 0)  Sleep(1);

		// Increment the counter
		++usCounter;
	}


	///////////
	// Cleanup
	///////////

	// Release the surface device context
	pDirectDrawSurface->ReleaseDC(hdc);

	// Release the DirectDraw palette if appropriate
	if (pDirectDrawPalette)  pDirectDrawPalette->Release();

	// Release the DirectDraw surface if appropriate
	if (pDirectDrawSurface)  pDirectDrawSurface->Release();

	// Release the DirectDraw object if appropriate
	if (pDirectDraw)		 pDirectDraw->Release();

	// Release memory allocated for the application icon and cursor images
	DestroyIcon(wcWindowClass.hIcon);
	DestroyCursor(wcWindowClass.hCursor);

	// Release memory allocated for the bitmap
	UnloadBitmap(&bsBitmap);

	return (msgMainWindow.wParam);
}
